import torch
import torch.nn as nn
import numpy as np
import torch.nn.functional as F

from .attack import Attack

from loss.RCE_loss import RCELoss
from loss.EFCE_nT_loss import EFCEnTLoss
from loss.EFCE_T_loss import EFCETLoss
from loss.EFRCE_loss import EFRCELoss

class BIM(Attack):
    r"""
    BIM or iterative-FGSM in the paper 'Adversarial Examples in the Physical World'
    [https://arxiv.org/abs/1607.02533]

    Distance Measure : Linf

    Arguments:
        model (nn.Module): model to attack.
        eps (float): maximum perturbation. (Default: 4/255)
        alpha (float): step size. (Default: 1/255)
        steps (int): number of steps. (Default: 0)

    .. note:: If steps set to 0, steps will be automatically decided following the paper.

    Shape:
        - images: :math:`(N, C, H, W)` where `N = number of batches`, `C = number of channels`,        `H = height` and `W = width`. It must have a range [0, 1].
        - labels: :math:`(N)` where each value :math:`y_i` is :math:`0 \leq y_i \leq` `number of labels`.
        - output: :math:`(N, C, H, W)`.

    Examples::
        >>> attack = torchattacks.BIM(model, eps=4/255, alpha=1/255, steps=0)
        >>> adv_images = attack(images, labels)
    """
    def __init__(self, model, eps=4/255, alpha=1/255, steps=0, loss_type='CE', temperature_scale=1.0, fuzzy_scale=1.0):
        super().__init__("BIM", model)
        self.eps = eps
        self.alpha = alpha
        if steps == 0:
            self.steps = int(min(eps*255 + 4, 1.25*eps*255))
        else:
            self.steps = steps
        self._supported_mode = ['default', 'targeted']
        self.loss_type = loss_type
        self.temperature_scale = temperature_scale
        self.fuzzy_scale = fuzzy_scale

    def forward(self, images, labels):
        r"""
        Overridden.
        """
        images = images.clone().detach().to(self.device)
        labels = labels.clone().detach().to(self.device)

        if self._targeted:
            target_labels = self._get_target_label(images, labels)


        if self.loss_type == 'CE':
            loss = nn.CrossEntropyLoss()
        elif self.loss_type == 'RCE':
            loss = RCELoss()
        elif self.loss_type == "EFCEnT":
            loss = EFCEnTLoss(temperature_scale=self.temperature_scale, fuzzy_scale=self.fuzzy_scale)
        elif self.loss_type == "EFCET":
            loss = EFCETLoss(temperature_scale=self.temperature_scale, fuzzy_scale=self.fuzzy_scale)
        elif self.loss_type == "EFRCE":
            loss = EFRCELoss(temperature_scale=self.temperature_scale, fuzzy_scale=self.fuzzy_scale)


        ori_images = images.clone().detach()

        for _ in range(self.steps):
            images.requires_grad = True
            outputs = self.model(images)

            # Calculate loss
            if self._targeted:
                cost = -loss(outputs, target_labels)
            else:
                cost = loss(outputs, labels)

            # Update adversarial images
            grad = torch.autograd.grad(cost, images,
                                       retain_graph=False,
                                       create_graph=False)[0]

            adv_images = images + self.alpha*grad.sign()
            a = torch.clamp(ori_images - self.eps, min=0)
            b = (adv_images >= a).float()*adv_images \
                + (adv_images < a).float()*a
            c = (b > ori_images+self.eps).float()*(ori_images+self.eps) \
                + (b <= ori_images + self.eps).float()*b
            images = torch.clamp(c, max=1).detach()

        if self._targeted:
            return images, target_labels
        else:
            return images

    def forward_proposition0(self, victim_model, images, labels):
        '''
        计算梯度夹角余弦值
        '''
        images = images.clone().detach().to(self.device)
        labels = labels.clone().detach().to(self.device)

        loss = nn.CrossEntropyLoss()

        ori_images = images.clone().detach()

        cosine_value_list = []
        for _ in range(self.steps):
            images.requires_grad = True
            outputs = self.model(images)

            # 计算替代模型梯度
            # Calculate loss
            cost = loss(outputs, labels)
            # Update adversarial images
            grad = torch.autograd.grad(cost, images,
                                       retain_graph=False,
                                       create_graph=False)[0]

            # 计算目标模型梯度
            outputs_victim = victim_model(images)
            # Calculate loss
            cost_victim = loss(outputs_victim, labels)
            # Get grad
            grad_victim = torch.autograd.grad(cost_victim, images,
                                              retain_graph=False,
                                              create_graph=False)[0]

            adv_images = images + self.alpha * grad.sign()
            a = torch.clamp(ori_images - self.eps, min=0)
            b = (adv_images >= a).float() * adv_images \
                + (adv_images < a).float() * a
            c = (b > ori_images + self.eps).float() * (ori_images + self.eps) \
                + (b <= ori_images + self.eps).float() * b
            images = torch.clamp(c, max=1).detach()

            grad = grad.view(grad.shape[0], -1)
            grad_victim = grad_victim.view(grad_victim.shape[0], -1)
            cosine_value = F.cosine_similarity(grad, grad_victim)
            cosine_value = cosine_value.detach().cpu().numpy()
            cosine_value = cosine_value.reshape(1,-1)

            cosine_value_list.append(cosine_value)
        cosine_value_list = np.concatenate(cosine_value_list)

        return cosine_value_list

    def forward_proposition1(self, images, labels):
        r"""
                验证proposition1.1和1.2
                """
        images = images.clone().detach().to(self.device)
        labels = labels.clone().detach().to(self.device)

        loss = nn.CrossEntropyLoss()

        ori_images = images.clone().detach()

        images_per_step = [images]

        for _ in range(self.steps):
            images.requires_grad = True
            outputs = self.model(images)

            # Calculate loss
            cost = loss(outputs, labels)

            # Update adversarial images
            grad = torch.autograd.grad(cost, images,
                                       retain_graph=False,
                                       create_graph=False)[0]

            adv_images = images + self.alpha * grad.sign()
            a = torch.clamp(ori_images - self.eps, min=0)
            b = (adv_images >= a).float() * adv_images \
                + (adv_images < a).float() * a
            c = (b > ori_images + self.eps).float() * (ori_images + self.eps) \
                + (b <= ori_images + self.eps).float() * b
            images = torch.clamp(c, max=1).detach()

            images_per_step.append(images)

        return images_per_step, images