import torch
import torch.nn as nn
import numpy as np
import torch.nn.functional as F

from .attack import Attack

from loss.RCE_loss import RCELoss
from loss.EFCE_nT_loss import EFCEnTLoss
from loss.EFCE_T_loss import EFCETLoss
from loss.EFRCE_loss import EFRCELoss

class FGSM(Attack):
    r"""
    FGSM in the paper 'Explaining and harnessing adversarial examples'
    [https://arxiv.org/abs/1412.6572]

    Distance Measure : Linf

    Arguments:
        model (nn.Module): model to attack.
        eps (float): maximum perturbation. (Default: 0.007)

    Shape:
        - images: :math:`(N, C, H, W)` where `N = number of batches`, `C = number of channels`,        `H = height` and `W = width`. It must have a range [0, 1].
        - labels: :math:`(N)` where each value :math:`y_i` is :math:`0 \leq y_i \leq` `number of labels`.
        - output: :math:`(N, C, H, W)`.

    Examples::
        >>> attack = torchattacks.FGSM(model, eps=0.007)
        >>> adv_images = attack(images, labels)

    """
    def __init__(self, model, eps=0.007, loss_type='CE', temperature_scale=1.0, fuzzy_scale=1.0):
        super().__init__("FGSM", model)
        self.eps = eps
        self._supported_mode = ['default', 'targeted']
        self.loss_type = loss_type
        self.temperature_scale = temperature_scale
        self.fuzzy_scale = fuzzy_scale

    def forward(self, images, labels):
        r"""
        Overridden.
        """
        images = images.clone().detach().to(self.device)
        labels = labels.clone().detach().to(self.device)

        if self._targeted:
            target_labels = self._get_target_label(images, labels)

        if self.loss_type == 'CE':
            loss = nn.CrossEntropyLoss()
        elif self.loss_type == 'RCE':
            loss = RCELoss()
        elif self.loss_type == 'EFCEnT':
            loss = EFCEnTLoss(temperature_scale=self.temperature_scale, fuzzy_scale=self.fuzzy_scale)
        elif self.loss_type == 'EFCET':
            loss = EFCETLoss(temperature_scale=self.temperature_scale, fuzzy_scale=self.fuzzy_scale)
        elif self.loss_type == 'EFRCE':
            loss = EFRCELoss(temperature_scale=self.temperature_scale, fuzzy_scale=self.fuzzy_scale)

        images.requires_grad = True
        outputs = self.model(images)

        # Calculate loss
        if self._targeted:
            cost = -loss(outputs, target_labels)
        else:
            cost = loss(outputs, labels)

        # Update adversarial images
        grad = torch.autograd.grad(cost, images,
                                   retain_graph=False, create_graph=False)[0]

        adv_images = images + self.eps*grad.sign()
        adv_images = torch.clamp(adv_images, min=0, max=1).detach()

        if self._targeted:
            return adv_images, target_labels
        else:
            return adv_images

    def forward_proposition0(self, victim_model, images, labels):
        '''
        验证proposition0的函数，返回梯度夹角列表
        '''
        images = images.clone().detach().to(self.device)
        labels = labels.clone().detach().to(self.device)

        loss = nn.CrossEntropyLoss()

        images.requires_grad = True
        # 计算替代模型产生的梯度
        outputs = self.model(images)
        # Calculate loss
        cost = loss(outputs, labels)
        # Get grad
        grad = torch.autograd.grad(cost, images,
                                   retain_graph=False, create_graph=False)[0]

        # 计算受害者模型产生的梯度
        outputs_victim = victim_model(images)
        # Calculate loss
        cost_victim = loss(outputs_victim, labels)
        # Get grad
        grad_victim = torch.autograd.grad(cost_victim, images,
                                   retain_graph=False, create_graph=False)[0]

        # 计算梯度夹角余弦值
        grad = grad.view(grad.shape[0], -1)
        grad_victim = grad_victim.view(grad_victim.shape[0], -1)
        cosine_value = F.cosine_similarity(grad, grad_victim)
        cosine_value = cosine_value.detach().cpu().numpy()

        return cosine_value