import torch
import torch.nn as nn
import numpy as np
import torch.nn.functional as F

from .attack import Attack

from loss.WACELoss import WACELoss
from loss.WA_EFCE_nT_loss import WAEFCEnTLoss

class QVMIFGSM(Attack):
    r"""
    QVMI-FGSM in the paper 'Enhancing the Transferability of Adversarial Attacks through Variance Tuning
    [https://arxiv.org/abs/2103.15571], Published as a conference paper at CVPR 2021
    Modified from "https://github.com/JHL-HUST/VT"

    Distance Measure : Linf

    Arguments:
        model (nn.Module): model to attack.
        eps (float): maximum perturbation. (Default: 8/255)
        alpha (float): step size. (Default: 2/255)
        decay (float): momentum factor. (Default: 1.0)
        steps (int): number of iterations. (Default: 5)
        N (int): the number of sampled examples in the neighborhood. (Default: 20)
        beta (float): the upper bound of neighborhood. (Default: 3/2)

    Shape:
        - images: :math:`(N, C, H, W)` where `N = number of batches`, `C = number of channels`,        `H = height` and `W = width`. It must have a range [0, 1].
        - labels: :math:`(N)` where each value :math:`y_i` is :math:`0 \leq y_i \leq` `number of labels`.
        - output: :math:`(N, C, H, W)`.

    Examples::
        >>> attack = torchattacks.QVMIFGSM(model, eps=8/255, steps=5, decay=1.0, N=20, beta=3/2)
        >>> adv_images = attack(images, labels)

    """

    def __init__(self, model, eps=8/255, alpha=2/255, steps=5, decay=1.0, N=20, beta=3/2, victim_model=None, num_classes=10, wtop_n=1, query_num=1, loss_type='WACE', temperature_scale=1.0, fuzzy_scale=1.0):
        super().__init__("QVMIFGSM", model)
        self.eps = eps
        self.steps = steps
        self.decay = decay
        self.alpha = alpha
        self.N = N
        self.beta = beta
        self._supported_mode = ['default', 'targeted']
        self.victim_model = victim_model
        self.num_classes = num_classes
        self.wtop_n = wtop_n
        self.query_num = query_num
        self.loss_type = loss_type
        self.temperature_scale = temperature_scale
        self.fuzzy_scale = fuzzy_scale

        # 根据self.steps和self.query_num确定需要查询victim model的位置，采取等距查询的方式
        assert self.query_num > 0 and self.query_num <= self.steps
        self.query_position = [int(np.floor(self.steps / self.query_num) * i) for i in range(self.query_num)]

    def forward(self, images, labels):
        r"""
        Overridden.
        """
        images = images.clone().detach().to(self.device)
        labels = labels.clone().detach().to(self.device)

        if self._targeted:
            target_labels = self._get_target_label(images, labels)


        momentum = torch.zeros_like(images).detach().to(self.device)

        v = torch.zeros_like(images).detach().to(self.device)

        if self.loss_type == 'WACE':
            loss = WACELoss(topn=self.wtop_n, num_classes=self.num_classes)
        elif self.loss_type == 'WAEFCEnT':
            loss = WAEFCEnTLoss(topn=self.wtop_n, num_classes=self.num_classes, temperature_scale=self.temperature_scale, fuzzy_scale=self.fuzzy_scale)

        adv_images = images.clone().detach()

        outputs_victim = None
        for _ in range(self.steps):
            adv_images.requires_grad = True
            outputs = self.model(adv_images)

            # Calculate loss
            # 受害者模型查询
            if _ in self.query_position:
                with torch.no_grad():
                    outputs_victim = self.victim_model(adv_images)
            if self._targeted:
                cost = loss(outputs, labels, outputs_victim, targeted=True, target_labels=target_labels)
            else:
                cost = loss(outputs, labels, outputs_victim)

            # Update adversarial images
            adv_grad = torch.autograd.grad(cost, adv_images,
                                       retain_graph=False, create_graph=False)[0]

            grad = (adv_grad + v) / torch.mean(torch.abs(adv_grad + v), dim=(1,2,3), keepdim=True)
            grad = grad + momentum * self.decay
            momentum = grad

            # Calculate Gradient Variance
            GV_grad = torch.zeros_like(images).detach().to(self.device)
            for _ in range(self.N):
                neighbor_images = adv_images.detach() + \
                                  torch.randn_like(images).uniform_(-self.eps*self.beta, self.eps*self.beta)
                neighbor_images.requires_grad = True
                outputs = self.model(neighbor_images)

                # Calculate loss
                if self._targeted:
                    cost = loss(outputs, labels, outputs_victim, targeted=True, target_labels=target_labels)
                else:
                    cost = loss(outputs, labels, outputs_victim)

                GV_grad += torch.autograd.grad(cost, neighbor_images,
                                               retain_graph=False, create_graph=False)[0]
            # obtaining the gradient variance
            v = GV_grad / self.N - adv_grad

            adv_images = adv_images.detach() + self.alpha*grad.sign()
            delta = torch.clamp(adv_images - images, min=-self.eps, max=self.eps)
            adv_images = torch.clamp(images + delta, min=0, max=1).detach()

        if self._targeted:
            return adv_images, target_labels
        else:
            return adv_images

    def forward_proposition0(self, victim_model, images, labels):
        '''
        计算梯度夹角余弦值
        '''
        images = images.clone().detach().to(self.device)
        labels = labels.clone().detach().to(self.device)

        momentum = torch.zeros_like(images).detach().to(self.device)

        v = torch.zeros_like(images).detach().to(self.device)

        loss = WACELoss(topn=self.wtop_n, num_classes=self.num_classes)

        loss_CE = nn.CrossEntropyLoss()

        adv_images = images.clone().detach()

        cosine_value_list = []
        outputs_victim = None
        for _ in range(self.steps):
            adv_images.requires_grad = True
            outputs = self.model(adv_images)

            # Calculate loss
            # 受害者模型查询
            if _ in self.query_position:
                with torch.no_grad():
                    outputs_victim = self.victim_model(adv_images)
            cost = loss(outputs, labels, outputs_victim)
            # Update adversarial images
            adv_grad = torch.autograd.grad(cost, adv_images,
                                           retain_graph=False, create_graph=False)[0]

            grad = (adv_grad + v) / torch.mean(torch.abs(adv_grad + v), dim=(1, 2, 3), keepdim=True)
            grad = grad + momentum * self.decay
            momentum = grad

            # Calculate Gradient Variance
            GV_grad = torch.zeros_like(images).detach().to(self.device)
            for _ in range(self.N):
                neighbor_images = adv_images.detach() + \
                                  torch.randn_like(images).uniform_(-self.eps * self.beta, self.eps * self.beta)
                neighbor_images.requires_grad = True
                outputs = self.model(neighbor_images)

                # Calculate loss
                cost = loss(outputs, labels, outputs_victim)
                GV_grad += torch.autograd.grad(cost, neighbor_images,
                                               retain_graph=False, create_graph=False)[0]
            # obtaining the gradient variance
            v = GV_grad / self.N - adv_grad

            # 计算目标模型梯度
            outputs_victim_temp = victim_model(adv_images)
            # calculate loss
            cost_victim = loss_CE(outputs_victim_temp, labels)
            # Get grad
            grad_victim = torch.autograd.grad(cost_victim, adv_images,
                                              retain_graph=False,
                                              create_graph=False)[0]

            adv_images = adv_images.detach() + self.alpha * grad.sign()
            delta = torch.clamp(adv_images - images, min=-self.eps, max=self.eps)
            adv_images = torch.clamp(images + delta, min=0, max=1).detach()

            grad = grad.view(grad.shape[0], -1)
            grad_victim = grad_victim.view(grad_victim.shape[0], -1)
            cosine_value = F.cosine_similarity(grad, grad_victim)
            cosine_value = cosine_value.detach().cpu().numpy()
            cosine_value = cosine_value.reshape(1, -1)

            cosine_value_list.append(cosine_value)
        cosine_value_list = np.concatenate(cosine_value_list)

        return cosine_value_list